#
#   Function
#   Copyright © 2024 NatML Inc. All Rights Reserved.
#

from ctypes import c_int

class FXNDtype(c_int):
    NULL = 0
    FLOAT16 = 1
    FLOAT32 = 2
    FLOAT64 = 3
    INT8 = 4
    INT16 = 5
    INT32 = 6
    INT64 = 7
    UINT8 = 8
    UINT16 = 9
    UINT32 = 10
    UINT64 = 11
    BOOL = 12
    STRING = 13
    LIST = 14
    DICT = 15
    IMAGE = 16
    BINARY = 17


#
#   Function
#   Copyright © 2024 NatML Inc. All Rights Reserved.
#

from ctypes import CDLL
from pathlib import Path
from .configuration import _register_fxn_configuration
from .prediction import _register_fxn_prediction
from .stream import _register_fxn_prediction_stream
from .predictor import _register_fxn_predictor
from .value import _register_fxn_value
from .map import _register_fxn_value_map
from .version import _register_fxn_version

def load_fxnc (path: Path) -> CDLL:
    # Open
    fxnc = CDLL(str(path))
    # Register
    fxnc = _register_fxn_value(fxnc)
    fxnc = _register_fxn_value_map(fxnc)
    fxnc = _register_fxn_configuration(fxnc)
    fxnc = _register_fxn_prediction(fxnc)
    fxnc = _register_fxn_prediction_stream(fxnc)
    fxnc = _register_fxn_predictor(fxnc)
    fxnc = _register_fxn_version(fxnc)
    # Return
    return fxnc


# 
#   Function
#   Copyright © 2024 NatML Inc. All Rights Reserved.
#

from typing import List

from ..api import GraphClient
from ..types import Predictor, PredictorStatus
from .user import PROFILE_FIELDS

class PredictorService:

    def __init__ (self, client: GraphClient) -> None:
        self.client = client

    def retrieve (self, tag: str) -> Predictor:
        """
        Retrieve a predictor.

        Parameters:
            tag (str): Predictor tag.

        Returns:
            Predictor: Predictor.
        """
        # Query
        response = self.client.query(f"""
            query ($input: PredictorInput!) {{
                predictor (input: $input) {{
                    {PREDICTOR_FIELDS}
                }}
            }}
            """,
            { "input": { "tag": tag } }
        )
        # Create predictor
        predictor = response["predictor"]
        predictor = Predictor(**predictor) if predictor else None
        # Return
        return predictor

    def list (
        self,
        owner: str=None,
        status: PredictorStatus=None,
        offset: int=None,
        count: int=None
    ) -> List[Predictor]:
        """
        List the current user's predictors.

        Parameters:
            owner (str): Predictor owner. This defaults to the current user.
            status (PredictorStatus): Predictor status. This defaults to `ACTIVE`.
            offset (int): Pagination offset.
            count (int): Pagination count.

        Returns:
            list: User predictors.
        """
        # Query
        response = self.client.query(f"""
            query ($user: UserInput, $predictors: UserPredictorsInput) {{
                user (input: $user) {{
                    predictors (input: $predictors) {{
                        {PREDICTOR_FIELDS}
                    }}
                }}
            }}
            """,
            {
                "user": { "username": owner } if owner else None,
                "predictors": { "status": status, "offset": offset, "count": count }
            }
        )
        # Check
        user = response["user"]
        if not user:
            return None
        # Create predictors
        predictors = response["user"]["predictors"]
        predictors = [Predictor(**predictor) for predictor in predictors]
        # Return
        return predictors

    def search (
        self,
        query: str,
        offset: int=None,
        count: int=None
    ) -> List[Predictor]:
        """
        Search predictors.

        Parameters:
            q (str): Search query.
            offset (int): Pagination offset.
            count (int): Pagination count.

        Returns:
            list: Relevant predictors.
        """
        # Query
        response = self.client.query(f"""
            query ($input: PredictorsInput!) {{
                predictors (input: $input) {{
                    {PREDICTOR_FIELDS}
                }}
            }}
            """,
            { "input": { "query": query, "offset": offset, "count": count } }
        )
        # Create predictors
        predictors = response["predictors"]
        predictors = [Predictor(**predictor) for predictor in predictors]
        # Return
        return predictors
    
    def delete (self, tag: str) -> bool:
        """
        Delete a predictor.

        Parameters:
            tag (str): Predictor tag.

        Returns:
            bool: Whether the predictor was successfully deleted.
        """
        # Query
        response = self.client.query(f"""
            mutation ($input: DeletePredictorInput!) {{
                deletePredictor (input: $input)
            }}
            """,
            { "input": { "tag": tag } }
        )
        # Return
        result = response["deletePredictor"]
        return result
    
    def archive (self, tag: str) -> Predictor:
        """
        Archive an active predictor.

        Parameters:
            tag (str): Predictor tag.

        Returns:
            Predictor: Archived predictor.
        """
        # Query
        response = self.client.query(f"""
            mutation ($input: ArchivePredictorInput!) {{
                archivePredictor (input: $input) {{
                    {PREDICTOR_FIELDS}
                }}
            }}
            """,
            { "input": { "tag": tag } }
        )
        # Create predictor
        predictor = response["archivePredictor"]
        predictor = Predictor(**predictor) if predictor else None
        # Return
        return predictor

    
PREDICTOR_FIELDS = f"""
tag
owner {{
    {PROFILE_FIELDS}
}}
name
status
access
created
description
card
media
signature {{
    inputs {{
        name
        type
        description
        range
        optional
        enumeration {{
            name
            value
        }}
        default_value: defaultValue {{
            data
            type
            shape
        }}
        schema
    }}
    outputs {{
        name
        type
        description
    }}
}}
license
"""


# 
#   Function
#   Copyright © 2024 NatML Inc. All Rights Reserved.
#

from rich import print_json
from rich.progress import Progress, SpinnerColumn, TextColumn
from typer import Argument, Option

from ..function import Function
from ..types import PredictorStatus
from .auth import get_access_key

def retrieve_predictor (
    tag: str=Argument(..., help="Predictor tag.")
):
    fxn = Function(get_access_key())
    predictor = fxn.predictors.retrieve(tag)
    predictor = predictor.model_dump() if predictor else None
    print_json(data=predictor)

def list_predictors (
    owner: str=Option(None, help="Predictor owner. This defaults to the current user."),
    status: PredictorStatus=Option(PredictorStatus.Active, help="Predictor status. This defaults to `ACTIVE`."),
    offset: int=Option(None, help="Pagination offset."),
    count: int=Option(None, help="Pagination count.")
):
    fxn = Function(get_access_key())
    predictors = fxn.predictors.list(
        owner=owner,
        status=status,
        offset=offset,
        count=count
    )
    predictors = [predictor.model_dump() for predictor in predictors] if predictors is not None else None
    print_json(data=predictors)

def search_predictors (
    query: str=Argument(..., help="Search query."),
    offset: int=Option(None, help="Pagination offset."),
    count: int=Option(None, help="Pagination count.")
):
    fxn = Function(get_access_key())
    predictors = fxn.predictors.search(query=query, offset=offset, count=count)
    predictors = [predictor.model_dump() for predictor in predictors]
    print_json(data=predictors)

def delete_predictor (
    tag: str=Argument(..., help="Predictor tag.")
):
    with Progress(
        SpinnerColumn(spinner_name="dots"),
        TextColumn("[progress.description]{task.description}"),
        transient=True
    ) as progress:
        progress.add_task(description="Deleting Function...", total=None)
        fxn = Function(get_access_key())
        result = fxn.predictors.delete(tag)
        print_json(data=result)

def archive_predictor (
    tag: str=Argument(..., help="Predictor tag.")
):
    fxn = Function(get_access_key())
    predictor = fxn.predictors.archive(tag)
    print_json(data=predictor.model_dump())


#
#   Function
#   Copyright © 2024 NatML Inc. All Rights Reserved.
#

from ctypes import c_char_p, c_int32, CDLL, POINTER, Structure
from .status import FXNStatus
from .value import FXNValueRef

class FXNValueMap(Structure): pass

FXNValueMapRef = POINTER(FXNValueMap)

def _register_fxn_value_map (fxnc: CDLL) -> CDLL:
    # FXNValueMapCreate
    fxnc.FXNValueMapCreate.argtypes = [POINTER(FXNValueMapRef)]
    fxnc.FXNValueMapCreate.restype = FXNStatus
    # FXNValueMapRelease
    fxnc.FXNValueMapRelease.argtypes = [FXNValueMapRef]
    fxnc.FXNValueMapRelease.restype = FXNStatus
    # FXNValueMapGetSize
    fxnc.FXNValueMapGetSize.argtypes = [FXNValueMapRef, POINTER(c_int32)]
    fxnc.FXNValueMapGetSize.restype = FXNStatus
    # FXNValueMapGetKey
    fxnc.FXNValueMapGetKey.argtypes = [FXNValueMapRef, c_int32, c_char_p, c_int32]
    fxnc.FXNValueMapGetKey.restype = FXNStatus
    # FXNValueMapGetValue
    fxnc.FXNValueMapGetValue.argtypes = [FXNValueMapRef, c_char_p, POINTER(FXNValueRef)]
    fxnc.FXNValueMapGetValue.restype = FXNStatus
    # FXNValueMapSetValue
    fxnc.FXNValueMapSetValue.argtypes = [FXNValueMapRef, c_char_p, FXNValueRef]
    fxnc.FXNValueMapSetValue.restype = FXNStatus
    # Return
    return fxnc


#
#   Function
#   Copyright © 2024 NatML Inc. All Rights Reserved.
#

from ctypes import c_char_p, c_int, c_int32, c_void_p, CDLL, POINTER, Structure
from .dtype import FXNDtype
from .status import FXNStatus

class FXNValueFlags(c_int):
    NONE = 0
    COPY_DATA = 1

class FXNValue(Structure): pass

FXNValueRef = POINTER(FXNValue)

def _register_fxn_value (fxnc: CDLL) -> CDLL:
    # FXNValueRelease
    fxnc.FXNValueRelease.argtypes = [FXNValueRef]
    fxnc.FXNValueRelease.restype = FXNStatus
    # FXNValueGetData
    fxnc.FXNValueGetData.argtypes = [FXNValueRef, POINTER(c_void_p)]
    fxnc.FXNValueGetData.restype = FXNStatus
    # FXNValueGetType
    fxnc.FXNValueGetType.argtypes = [FXNValueRef, POINTER(FXNDtype)]
    fxnc.FXNValueGetType.restype = FXNStatus
    # FXNValueGetDimensions
    fxnc.FXNValueGetDimensions.argtypes = [FXNValueRef, POINTER(c_int32)]
    fxnc.FXNValueGetDimensions.restype = FXNStatus
    # FXNValueGetShape
    fxnc.FXNValueGetShape.argtypes = [FXNValueRef, POINTER(c_int32), c_int32]
    fxnc.FXNValueGetShape.restype = FXNStatus
    # FXNValueCreateArray
    fxnc.FXNValueCreateArray.argtypes = [c_void_p, POINTER(c_int32), c_int32, FXNDtype, FXNValueFlags, POINTER(FXNValueRef)]
    fxnc.FXNValueCreateArray.restype = FXNStatus
    # FXNValueCreateString
    fxnc.FXNValueCreateString.argtypes = [c_char_p, POINTER(FXNValueRef)]
    fxnc.FXNValueCreateString.restype = FXNStatus
    # FXNValueCreateList
    fxnc.FXNValueCreateList.argtypes = [c_char_p, POINTER(FXNValueRef)]
    fxnc.FXNValueCreateList.restype = FXNStatus
    # FXNValueCreateDict
    fxnc.FXNValueCreateDict.argtypes = [c_char_p, POINTER(FXNValueRef)]
    fxnc.FXNValueCreateDict.restype = FXNStatus
    # FXNValueCreateImage
    fxnc.FXNValueCreateImage.argtypes = [c_void_p, c_int32, c_int32, c_int32, FXNValueFlags, POINTER(FXNValueRef)]
    fxnc.FXNValueCreateImage.restype = FXNStatus
    # Return
    return fxnc


#
#   Function
#   Copyright © 2024 NatML Inc. All Rights Reserved.
#

from ctypes import c_char_p, c_double, c_int32, CDLL, POINTER, Structure
from .status import FXNStatus
from .map import FXNValueMapRef

class FXNPrediction(Structure): pass

FXNPredictionRef = POINTER(FXNPrediction)

def _register_fxn_prediction (fxnc: CDLL) -> CDLL:
    # FXNPredictionRelease
    fxnc.FXNPredictionRelease.argtypes = [FXNPredictionRef]
    fxnc.FXNPredictionRelease.restype = FXNStatus
    # FXNPredictionGetID
    fxnc.FXNPredictionGetID.argtypes = [FXNPredictionRef, c_char_p, c_int32]
    fxnc.FXNPredictionGetID.restype = FXNStatus
    # FXNPredictionGetLatency
    fxnc.FXNPredictionGetLatency.argtypes = [FXNPredictionRef, POINTER(c_double)]
    fxnc.FXNPredictionGetLatency.restype = FXNStatus
    # FXNPredictionGetResults
    fxnc.FXNPredictionGetResults.argtypes = [FXNPredictionRef, POINTER(FXNValueMapRef)]
    fxnc.FXNPredictionGetResults.restype = FXNStatus
    # FXNPredictionGetError
    fxnc.FXNPredictionGetError.argtypes = [FXNPredictionRef, c_char_p, c_int32]
    fxnc.FXNPredictionGetError.restype = FXNStatus
    # FXNPredictionGetLogs
    fxnc.FXNPredictionGetLogs.argtypes = [FXNPredictionRef, c_char_p, c_int32]
    fxnc.FXNPredictionGetLogs.restype = FXNStatus
    # FXNPredictionGetLogLength
    fxnc.FXNPredictionGetLogLength.argtypes = [FXNPredictionRef, POINTER(c_int32)]
    fxnc.FXNPredictionGetLogLength.restype = FXNStatus
    # Return
    return fxnc


#
#   Function
#   Copyright © 2024 NatML Inc. All Rights Reserved.
#

from ctypes import c_char_p, c_int, c_int32, c_void_p, CDLL, POINTER, Structure
from .status import FXNStatus

class FXNAcceleration(c_int):
    FXN_ACCELERATION_DEFAULT = 0
    FXN_ACCELERATION_CPU = 1 << 0
    FXN_ACCELERATION_GPU = 1 << 1
    FXN_ACCELERATION_NPU = 1 << 2

class FXNConfiguration(Structure): pass

FXNConfigurationRef = POINTER(FXNConfiguration)

def _register_fxn_configuration (fxnc: CDLL) -> CDLL:
    # FXNConfigurationGetUniqueID
    fxnc.FXNConfigurationGetUniqueID.argtypes = [c_char_p, c_int32]
    fxnc.FXNConfigurationGetUniqueID.restype = FXNStatus
    # FXNConfigurationGetClientID
    fxnc.FXNConfigurationGetClientID.argtypes = [c_char_p, c_int32]
    fxnc.FXNConfigurationGetClientID.restype = FXNStatus
    # FXNConfigurationCreate
    fxnc.FXNConfigurationCreate.argtypes = [POINTER(FXNConfigurationRef)]
    fxnc.FXNConfigurationCreate.restype = FXNStatus
    # FXNConfigurationRelease
    fxnc.FXNConfigurationRelease.argtypes = [FXNConfigurationRef]
    fxnc.FXNConfigurationRelease.restype = FXNStatus
    # FXNConfigurationGetTag
    fxnc.FXNConfigurationGetTag.argtypes = [FXNConfigurationRef, c_char_p, c_int32]
    fxnc.FXNConfigurationRelease.restype = FXNStatus
    # FXNConfigurationSetTag
    fxnc.FXNConfigurationSetTag.argtypes = [FXNConfigurationRef, c_char_p]
    fxnc.FXNConfigurationSetTag.restype = FXNStatus
    # FXNConfigurationGetToken
    fxnc.FXNConfigurationGetToken.argtypes = [FXNConfigurationRef, c_char_p, c_int32]
    fxnc.FXNConfigurationGetToken.restype = FXNStatus
    # FXNConfigurationSetToken
    fxnc.FXNConfigurationSetToken.argtypes = [FXNConfigurationRef, c_char_p]
    fxnc.FXNConfigurationSetToken.restype = FXNStatus
    # FXNConfigurationGetAcceleration
    fxnc.FXNConfigurationGetAcceleration.argtypes = [FXNConfigurationRef, POINTER(FXNAcceleration)]
    fxnc.FXNConfigurationGetAcceleration.restype = FXNStatus
    # FXNConfigurationSetAcceleration
    fxnc.FXNConfigurationSetAcceleration.argtypes = [FXNConfigurationRef, FXNAcceleration]
    fxnc.FXNConfigurationSetAcceleration.restype = FXNStatus
    # FXNConfigurationGetDevice
    fxnc.FXNConfigurationGetDevice.argtypes = [FXNConfigurationRef, POINTER(c_void_p)]
    fxnc.FXNConfigurationGetDevice.restype = FXNStatus
    # FXNConfigurationSetDevice
    fxnc.FXNConfigurationSetDevice.argtypes = [FXNConfigurationRef, c_void_p]
    fxnc.FXNConfigurationSetDevice.restype = FXNStatus
    # FXNConfigurationAddResource
    fxnc.FXNConfigurationAddResource.argtypes = [FXNConfigurationRef, c_char_p, c_char_p]
    fxnc.FXNConfigurationAddResource.restype = FXNStatus
    # Return
    return fxnc


#
#   Function
#   Copyright © 2024 NatML Inc. All Rights Reserved.
#

# https://github.com/fxnai/fxnc

from .status import FXNStatus
from .value import FXNDtype, FXNValueRef, FXNValueFlags
from .map import FXNValueMapRef
from .configuration import FXNConfigurationRef, FXNAcceleration
from .prediction import FXNPredictionRef
from .stream import FXNPredictionStreamRef
from .predictor import FXNPredictorRef

from .fxnc import load_fxnc


# 
#   Function
#   Copyright © 2024 NatML Inc. All Rights Reserved.
#

from enum import Enum, IntFlag
from io import BytesIO
from numpy.typing import NDArray
from PIL import Image
from pydantic import AliasChoices, BaseModel, ConfigDict, Field
from typing import Any, Dict, List, Optional, Tuple

from .dtype import Dtype
from .user import Profile

class Acceleration (IntFlag):
    """
    Predictor acceleration.
    """
    Default = 0,
    CPU     = 1 << 0,
    GPU     = 1 << 1,
    NPU     = 1 << 2

class AccessMode (str, Enum):
    """
    Predictor access mode.
    """
    Public = "PUBLIC"
    Private = "PRIVATE"

class PredictorStatus (str, Enum):
    """
    Predictor status.
    """
    Provisioning = "PROVISIONING"
    Active = "ACTIVE"
    Invalid = "INVALID"
    Archived = "ARCHIVED"

class EnumerationMember (BaseModel):
    """
    Prediction parameter enumeration member.

    Members:
        name (str): Enumeration member name.
        value (str | int): Enumeration member value.
    """
    name: str = Field(description="Enumeration member name.")
    value: str | int = Field(description="Enumeration member value.")

class Parameter (BaseModel):
    """
    Predictor parameter.

    Members:
        name (str): Parameter name.
        type (Dtype): Parameter type. This is `None` if the type is unknown or unsupported by Function.
        description (str): Parameter description.
        optional (bool): Whether the parameter is optional.
        range (tuple): Parameter value range for numeric parameters.
        enumeration (list): Parameter value choices for enumeration parameters.
        default_value (str | float | int | bool | ndarray | list | dict | PIL.Image | BytesIO): Parameter default value.
        value_schema (dict): Parameter JSON schema. This is only populated for `list` and `dict` parameters.
    """
    name: str = Field(description="Parameter name.")
    type: Optional[Dtype] = Field(default=None, description="Parameter type. This is `None` if the type is unknown or unsupported by Function.")
    description: Optional[str] = Field(default=None, description="Parameter description.")
    optional: Optional[bool] = Field(default=None, description="Whether the parameter is optional.")
    range: Optional[Tuple[float, float]] = Field(default=None, description="Parameter value range for numeric parameters.")
    enumeration: Optional[List[EnumerationMember]] = Field(default=None, description="Parameter value choices for enumeration parameters.")
    default_value: Optional[str | float | int | bool | NDArray | List[Any] | Dict[str, Any] | Image.Image | BytesIO] = Field(default=None, description="Parameter default value.", serialization_alias="defaultValue", validation_alias=AliasChoices("default_value", "defaultValue"))
    value_schema: Optional[dict] = Field(default=None, description="Parameter JSON schema. This is only populated for `list` and `dict` parameters.", serialization_alias="schema", validation_alias=AliasChoices("schema", "value_schema"))
    model_config = ConfigDict(arbitrary_types_allowed=True)

class Signature (BaseModel):
    """
    Predictor signature.

    Members:
        inputs (list): Input parameters.
        outputs (list): Output parameters.
    """
    inputs: List[Parameter] = Field(description="Input parameters.")
    outputs: List[Parameter] = Field(description="Output parameters.")

class Predictor (BaseModel):
    """
    Predictor.

    Members:
        tag (str): Predictor tag.
        owner (Profile): Predictor owner.
        name (str): Predictor name.
        status (PredictorStatus): Predictor status.
        access (AccessMode): Predictor access.
        signature (Signature): Predictor signature.
        created (str): Date created.
        description (str): Predictor description.
        card (str): Predictor card.
        media (str): Predictor media URL.
        license (str): Predictor license URL.
    """
    tag: str = Field(description="Predictor tag.")
    owner: Profile = Field(description="Predictor owner.")
    name: str = Field(description="Predictor name.")
    status: PredictorStatus = Field(description="Predictor status.")
    access: AccessMode = Field(description="Predictor access.")
    signature: Signature = Field(description="Predictor signature.")
    created: str = Field(description="Date created.")
    description: Optional[str] = Field(default=None, description="Predictor description.")
    card: Optional[str] = Field(default=None, description="Predictor card.")
    media: Optional[str] = Field(default=None, description="Predictor media URL.")
    license: Optional[str] = Field(default=None, description="Predictor license URL.")


#
#   Function
#   Copyright © 2024 NatML Inc. All Rights Reserved.
#

from ctypes import c_int

class FXNStatus(c_int):
    OK = 0
    ERROR_INVALID_ARGUMENT = 1
    ERROR_INVALID_OPERATION = 2
    ERROR_NOT_IMPLEMENTED = 3


#
#   Function
#   Copyright © 2024 NatML Inc. All Rights Reserved.
#

from ctypes import CDLL, POINTER, Structure
from .prediction import FXNPredictionRef
from .status import FXNStatus

class FXNPredictionStream(Structure): pass

FXNPredictionStreamRef = POINTER(FXNPredictionStream)

def _register_fxn_prediction_stream (fxnc: CDLL) -> CDLL:
    # FXNPredictionStreamRelease
    fxnc.FXNPredictionStreamRelease.argtypes = [FXNPredictionStreamRef]
    fxnc.FXNPredictionStreamRelease.restype = FXNStatus
    # FXNPredictionStreamReadNext
    fxnc.FXNPredictionStreamReadNext.argtypes = [FXNPredictionStreamRef, POINTER(FXNPredictionRef)]
    fxnc.FXNPredictionStreamReadNext.restype = FXNStatus
    # Return
    return fxnc


#
#   Function
#   Copyright © 2024 NatML Inc. All Rights Reserved.
#

from ctypes import c_char_p, CDLL

def _register_fxn_version (fxnc: CDLL) -> CDLL:
    # FXNGetVersion
    fxnc.FXNGetVersion.argtypes = []
    fxnc.FXNGetVersion.restype = c_char_p
    # Return
    return fxnc


# 
#   Function
#   Copyright © 2024 NatML Inc. All Rights Reserved.
#

from .dtype import Dtype
from .prediction import Prediction, PredictionResource
from .predictor import Acceleration, AccessMode, EnumerationMember, Parameter, Predictor, PredictorStatus, Signature
from .user import Profile, User


#
#   Function
#   Copyright © 2024 NatML Inc. All Rights Reserved.
#

from ctypes import CDLL, POINTER, Structure
from .configuration import FXNConfigurationRef
from .prediction import FXNPredictionRef
from .stream import FXNPredictionStreamRef
from .status import FXNStatus
from .map import FXNValueMapRef

class FXNPredictor(Structure): pass

FXNPredictorRef = POINTER(FXNPredictor)

def _register_fxn_predictor (fxnc: CDLL) -> CDLL:
    # FXNPredictorCreate
    fxnc.FXNPredictorCreate.argtypes = [FXNConfigurationRef, POINTER(FXNPredictorRef)]
    fxnc.FXNPredictorCreate.restype = FXNStatus
    # FXNPredictorRelease
    fxnc.FXNPredictorRelease.argtypes = [FXNPredictorRef]
    fxnc.FXNPredictorRelease.restype = FXNStatus
    # FXNPredictorCreatePrediction
    fxnc.FXNPredictorCreatePrediction.argtypes = [FXNPredictorRef, FXNValueMapRef, POINTER(FXNPredictionRef)]
    fxnc.FXNPredictorCreatePrediction.restype = FXNStatus
    # FXNPredictorStreamPrediction
    fxnc.FXNPredictorStreamPrediction.argtypes = [FXNPredictionRef, FXNValueMapRef, POINTER(FXNPredictionStreamRef)]
    fxnc.FXNPredictorStreamPrediction.restype = FXNStatus
    # Return
    return fxnc


#
#   Function
#   Copyright © 2024 NatML Inc. All Rights Reserved.
#

from ctypes import byref, cast, c_char_p, c_double, c_int32, c_uint8, c_void_p, create_string_buffer, string_at, CDLL, POINTER
from dataclasses import asdict, is_dataclass
from datetime import datetime, timezone
from importlib import resources
from io import BytesIO
from json import dumps, loads
from numpy import array, dtype, int32, ndarray, zeros
from numpy.ctypeslib import as_array, as_ctypes_type
from numpy.typing import NDArray
from pathlib import Path
from PIL import Image
from platform import machine, system
from pydantic import BaseModel
from requests import get, post
from tempfile import gettempdir
from typing import Any, AsyncIterator, Dict, List, Optional, Union
from urllib.parse import urlparse

from ..api import GraphClient
from ..c import load_fxnc, FXNConfigurationRef, FXNDtype, FXNPredictionRef, FXNPredictorRef, FXNStatus, FXNValueRef, FXNValueFlags, FXNValueMapRef
from ..types import Acceleration, Prediction, PredictionResource

class PredictionService:

    def __init__ (self, client: GraphClient):
        self.client = client
        self.__fxnc = PredictionService.__load_fxnc()
        self.__cache = { }

    def create (
        self,
        tag: str,
        *,
        inputs: Optional[Dict[str, ndarray | str | float | int | bool | List[Any] | Dict[str, Any] | Path | Image.Image]] = None,
        acceleration: Acceleration=Acceleration.Default,
        client_id: str=None,
        configuration_id: str=None
    ) -> Prediction:
        """
        Create a prediction.

        Parameters:
            tag (str): Predictor tag.
            inputs (dict): Input values.
            acceleration (Acceleration): Prediction acceleration.
            client_id (str): Function client identifier. Specify this to override the current client identifier.
            configuration_id (str): Configuration identifier. Specify this to override the current client configuration identifier.

        Returns:
            Prediction: Created prediction.
        """
        # Check if cached
        if tag in self.__cache:
            return self.__predict(tag=tag, predictor=self.__cache[tag], inputs=inputs)
        # Query
        response = post(
            f"{self.client.api_url}/predict/{tag}",
            json={ },
            headers={
                "Authorization": f"Bearer {self.client.access_key}",
                "fxn-client": client_id if client_id is not None else self.__get_client_id(),
                "fxn-configuration-token": configuration_id if configuration_id is not None else self.__get_configuration_id()
            }
        )
        # Check
        prediction = response.json()
        try:
            response.raise_for_status()
        except Exception as ex:
            error = prediction["errors"][0]["message"] if "errors" in prediction else str(ex)
            raise RuntimeError(error)
        # Check raw prediction
        prediction = Prediction(**prediction)
        if inputs is None:
            return prediction
        # Create edge prediction
        predictor = self.__load(prediction, acceleration=acceleration)
        self.__cache[tag] = predictor
        prediction = self.__predict(tag=tag, predictor=predictor, inputs=inputs)
        # Return
        return prediction
    
    async def stream ( # INCOMPLETE # Streaming support
        self,
        tag: str,
        *,
        inputs: Dict[str, Union[float, int, str, bool, NDArray, List[Any], Dict[str, Any], Path, Image.Image]] = {},
        acceleration: Acceleration=Acceleration.Default,
        client_id: str=None,
        configuration_id: str=None
    ) -> AsyncIterator[Prediction]:
        """
        Create a streaming prediction.

        NOTE: This feature is currently experimental.

        Parameters:
            tag (str): Predictor tag.
            inputs (dict): Input values.
            acceleration (Acceleration): Prediction acceleration.
            client_id (str): Function client identifier. Specify this to override the current client identifier.
            configuration_id (str): Configuration identifier. Specify this to override the current client configuration identifier.

        Returns:
            Prediction: Created prediction.
        """
        # Check if cached
        if tag in self.__cache:
            yield self.__predict(tag=tag, predictor=self.__cache[tag], inputs=inputs)
            return
        # Create prediction
        prediction = self.create(
            tag=tag,
            client_id=client_id,
            configuration_id=configuration_id
        )
        # Make single prediction
        predictor = self.__load(prediction, acceleration=acceleration)
        self.__cache[tag] = predictor
        prediction = self.__predict(tag=tag, predictor=predictor, inputs=inputs)
        # Yield
        yield prediction

    @classmethod
    def __load_fxnc (self) -> Optional[CDLL]:
        os = system().lower()
        os = "macos" if os == "darwin" else os
        arch = machine()
        arch = "arm64" if arch == "aarch64" else arch
        arch = "x86_64" if arch == "x64" else arch
        package = f"fxn.lib.{os}.{arch}"
        resource = "libFunction.so"
        resource = "Function.dylib" if os == "macos" else resource
        resource = "Function.dll" if os == "windows" else resource
        with resources.path(package, resource) as fxnc_path:
            return load_fxnc(fxnc_path)

    def __get_client_id (self) -> str:
        # Fallback if fxnc failed to load
        if not self.__fxnc:
            os = system().lower()
            os = "macos" if os == "darwin" else os
            return f"{os}-{machine()}"
        # Get
        buffer = create_string_buffer(64)
        status = self.__fxnc.FXNConfigurationGetClientID(buffer, len(buffer))
        assert status.value == FXNStatus.OK, \
            f"Failed to retrieve prediction client identifier with status: {status.value}"
        client_id = buffer.value.decode("utf-8")
        # Return
        return client_id

    def __get_configuration_id (self) -> Optional[str]:
        # Check
        if not self.__fxnc:
            return None
        # Get
        buffer = create_string_buffer(2048)
        status = self.__fxnc.FXNConfigurationGetUniqueID(buffer, len(buffer))
        assert status.value == FXNStatus.OK, \
            f"Failed to retrieve prediction configuration identifier with error: {self.__class__.__status_to_error(status.value)}"
        uid = buffer.value.decode("utf-8")
        # Return
        return uid

    def __load (
        self,
        prediction: Prediction,
        *,
        acceleration: Acceleration=Acceleration.Default
    ) -> type[FXNPredictorRef]:
        fxnc = self.__fxnc
        configuration = FXNConfigurationRef()
        try:
            # Create configuration
            status = fxnc.FXNConfigurationCreate(byref(configuration))
            assert status.value == FXNStatus.OK, \
                f"Failed to create {prediction.tag} configuration with error: {self.__class__.__status_to_error(status.value)}"
            status = fxnc.FXNConfigurationSetTag(configuration, prediction.tag.encode())
            assert status.value == FXNStatus.OK, \
                f"Failed to set configuration tag with error: {self.__class__.__status_to_error(status.value)}"
            status = fxnc.FXNConfigurationSetToken(configuration, prediction.configuration.encode())
            assert status.value == FXNStatus.OK, \
                f"Failed to set configuration token with error: {self.__class__.__status_to_error(status.value)}"
            status = fxnc.FXNConfigurationSetAcceleration(configuration, int(acceleration))
            assert status.value == FXNStatus.OK, \
                f"Failed to set configuration acceleration with error: {self.__class__.__status_to_error(status.value)}"
            for resource in prediction.resources:
                if resource.type == "fxn": # CHECK # Remove in fxnc 0.0.27
                    continue
                path = self.__get_resource_path(resource)
                status = fxnc.FXNConfigurationAddResource(configuration, resource.type.encode(), str(path).encode())
                assert status.value == FXNStatus.OK, \
                    f"Failed to set prediction configuration resource with type {resource.type} for tag {prediction.tag} with error: {self.__class__.__status_to_error(status.value)}"
            # Create predictor
            predictor = FXNPredictorRef()
            status = fxnc.FXNPredictorCreate(configuration, byref(predictor))
            assert status.value == FXNStatus.OK, \
                f"Failed to create prediction for tag {prediction.tag} with error: {self.__class__.__status_to_error(status.value)}"
            # Return
            return predictor
        finally:
            fxnc.FXNConfigurationRelease(configuration)

    def __predict (self, *, tag: str, predictor, inputs: Dict[str, Any]) -> Prediction:
        fxnc = self.__fxnc
        input_map = FXNValueMapRef()
        prediction = FXNPredictionRef()
        try:
            # Marshal inputs
            status = fxnc.FXNValueMapCreate(byref(input_map))
            assert status.value == FXNStatus.OK, \
                f"Failed to create {tag} prediction because input values could not be provided to the predictor with error: {self.__class__.__status_to_error(status.value)}"
            for name, value in inputs.items():
                value = self.__to_value(value)
                fxnc.FXNValueMapSetValue(input_map, name.encode(), value)
            # Predict
            status = fxnc.FXNPredictorCreatePrediction(predictor, input_map, byref(prediction))
            assert status.value == FXNStatus.OK, \
                f"Failed to create {tag} prediction with error: {self.__class__.__status_to_error(status.value)}"
            # Marshal prediction
            id = create_string_buffer(256)
            error = create_string_buffer(2048)
            latency = c_double()
            status = fxnc.FXNPredictionGetID(prediction, id, len(id))
            assert status.value == FXNStatus.OK, \
                f"Failed to get {tag} prediction identifier with error: {self.__class__.__status_to_error(status.value)}"
            status = fxnc.FXNPredictionGetLatency(prediction, byref(latency))
            assert status.value == FXNStatus.OK, \
                f"Failed to get {tag} prediction latency with error: {self.__class__.__status_to_error(status.value)}"
            fxnc.FXNPredictionGetError(prediction, error, len(error))
            id = id.value.decode("utf-8")
            latency = latency.value
            error = error.value.decode("utf-8")
            log_length = c_int32()
            fxnc.FXNPredictionGetLogLength(prediction, byref(log_length))
            logs = create_string_buffer(log_length.value + 1)
            fxnc.FXNPredictionGetLogs(prediction, logs, len(logs))
            logs = logs.value.decode("utf-8")
            # Marshal outputs
            results = []
            output_count = c_int32()
            output_map = FXNValueMapRef()
            status = fxnc.FXNPredictionGetResults(prediction, byref(output_map))
            assert status.value == FXNStatus.OK, f"Failed to get {tag} prediction results with error: {self.__class__.__status_to_error(status.value)}"
            status = fxnc.FXNValueMapGetSize(output_map, byref(output_count))
            assert status.value == FXNStatus.OK, f"Failed to get {tag} prediction result count with error: {self.__class__.__status_to_error(status.value)}"
            for idx in range(output_count.value):
                name = create_string_buffer(256)
                status = fxnc.FXNValueMapGetKey(output_map, idx, name, len(name))
                assert status.value == FXNStatus.OK, \
                    f"Failed to get {tag} prediction output name at index {idx} with error: {self.__class__.__status_to_error(status.value)}"
                value = FXNValueRef()
                status = fxnc.FXNValueMapGetValue(output_map, name, byref(value))
                assert status.value == FXNStatus.OK, \
                    f"Failed to get {tag} prediction output value at index {idx} with error: {self.__class__.__status_to_error(status.value)}"
                name = name.value.decode("utf-8")
                value = self.__to_object(value)
                results.append(value)
            # Return
            return Prediction(
                id=id,
                tag=tag,
                results=results if not error else None,
                latency=latency,
                error=error if error else None,
                logs=logs,
                created=datetime.now(timezone.utc).isoformat()
            )
        finally:
            fxnc.FXNPredictionRelease(prediction)
            fxnc.FXNValueMapRelease(input_map)
    
    def __to_value (
        self,
        value: Union[float, int, bool, str, NDArray, List[Any], Dict[str, Any], Image.Image, bytes, bytearray, memoryview, BytesIO, None]
    ) -> type[FXNValueRef]:
        value = PredictionService.__try_ensure_serializable(value)
        fxnc = self.__fxnc
        result = FXNValueRef()
        if result is None:
            fxnc.FXNValueCreateNull(byref(result))
        elif isinstance(value, bool):
            return self.__to_value(array(value, dtype="bool"))
        elif isinstance(value, int):
            return self.__to_value(array(value, dtype="int32"))
        elif isinstance(value, float):
            return self.__to_value(array(value, dtype="float32"))
        elif isinstance(value, ndarray):
            dtype = _NP_TO_FXN_DTYPE.get(value.dtype)
            assert dtype is not None, f"Failed to convert numpy array to Function value because array data type is not supported: {value.dtype}"
            fxnc.FXNValueCreateArray(
                value.ctypes.data_as(c_void_p),
                value.ctypes.shape_as(c_int32),
                len(value.shape),
                dtype,
                FXNValueFlags.NONE,
                byref(result)
            )
        elif isinstance(value, str):
            fxnc.FXNValueCreateString(value.encode(), byref(result))
        elif isinstance(value, list):
            fxnc.FXNValueCreateList(dumps(value).encode(), byref(result))
        elif isinstance(value, dict):
            fxnc.FXNValueCreateDict(dumps(value).encode(), byref(result))
        elif isinstance(value, Image.Image):
            value = array(value)
            status = fxnc.FXNValueCreateImage(
                value.ctypes.data_as(c_void_p),
                value.shape[1],
                value.shape[0],
                value.shape[2],
                FXNValueFlags.COPY_DATA,
                byref(result)
            )
            assert status.value == FXNStatus.OK, f"Failed to create image value with error: {self.__class__.__status_to_error(status.value)}"
        elif isinstance(value, (bytes, bytearray, memoryview, BytesIO)):
            copy = isinstance(value, memoryview) 
            view = memoryview(value.getvalue() if isinstance(value, BytesIO) else value) if not isinstance(value, memoryview) else value
            buffer = (c_uint8 * len(view)).from_buffer(view)
            fxnc.FXNValueCreateBinary(
                buffer,
                len(view),
                FXNValueFlags.COPY_DATA if copy else FXNValueFlags.NONE,
                byref(result)
            )
        else:
            raise RuntimeError(f"Failed to convert Python value to Function value because Python value has an unsupported type: {type(value)}")
        return result
    
    def __to_object (
        self,
        value: type[FXNValueRef]
    ) -> Union[float, int, bool, str, NDArray, List[Any], Dict[str, Any], Image.Image, BytesIO, None]:
        # Type
        fxnc = self.__fxnc
        dtype = FXNDtype()
        status = fxnc.FXNValueGetType(value, byref(dtype))
        assert status.value == FXNStatus.OK, f"Failed to get value data type with error: {self.__class__.__status_to_error(status.value)}"
        dtype = dtype.value
        # Get data
        data = c_void_p()
        status = fxnc.FXNValueGetData(value, byref(data))
        assert status.value == FXNStatus.OK, f"Failed to get value data with error: {self.__class__.__status_to_error(status.value)}"
        # Get shape
        dims = c_int32()
        status = fxnc.FXNValueGetDimensions(value, byref(dims))
        assert status.value == FXNStatus.OK, f"Failed to get value dimensions with error: {self.__class__.__status_to_error(status.value)}"
        shape = zeros(dims.value, dtype=int32)
        status = fxnc.FXNValueGetShape(value, shape.ctypes.data_as(POINTER(c_int32)), dims)
        assert status.value == FXNStatus.OK, f"Failed to get value shape with error: {self.__class__.__status_to_error(status.value)}"
        # Switch
        if dtype == FXNDtype.NULL:
            return None
        elif dtype in _FXN_TO_NP_DTYPE:
            dtype_c = as_ctypes_type(_FXN_TO_NP_DTYPE[dtype])
            tensor = as_array(cast(data, POINTER(dtype_c)), shape)
            return tensor.item() if len(tensor.shape) == 0 else tensor.copy()
        elif dtype == FXNDtype.STRING:
            return cast(data, c_char_p).value.decode()
        elif dtype == FXNDtype.LIST:
            return loads(cast(data, c_char_p).value.decode())
        elif dtype == FXNDtype.DICT:
            return loads(cast(data, c_char_p).value.decode())
        elif dtype == FXNDtype.IMAGE:
            pixel_buffer = as_array(cast(data, POINTER(c_uint8)), shape)
            return Image.fromarray(pixel_buffer.copy())
        elif dtype == FXNDtype.BINARY:
            return BytesIO(string_at(data, shape[0]))
        else:
            raise RuntimeError(f"Failed to convert Function value to Python value because Function value has unsupported type: {dtype}")

    def __get_resource_path (self, resource: PredictionResource) -> Path:
        cache_dir = self.__class__.__get_resource_dir() / ".fxn" / "cache"
        cache_dir.mkdir(exist_ok=True)
        res_name = Path(urlparse(resource.url).path).name
        res_path = cache_dir / res_name
        if res_path.exists():
            return res_path
        req = get(resource.url)
        req.raise_for_status()
        with open(res_path, "wb") as f:
            f.write(req.content)
        return res_path

    @classmethod
    def __get_resource_dir (cls) -> Path:
        try:
            check = Path.home() / ".fxntest"
            with open(check, "w") as f:
                f.write("fxn")
            check.unlink()
            return Path.home()
        except:
            return Path(gettempdir())

    @classmethod
    def __try_ensure_serializable (cls, object: Any) -> Any:
        if object is None:
            return object
        if isinstance(object, list):
            return [cls.__try_ensure_serializable(x) for x in object]
        if is_dataclass(object) and not isinstance(object, type):
            return asdict(object)
        if isinstance(object, BaseModel):
            return object.model_dump(mode="json", by_alias=True)
        return object
    
    @classmethod
    def __status_to_error (cls, status: int) -> str:
        if status == FXNStatus.ERROR_INVALID_ARGUMENT:
            return "FXN_ERROR_INVALID_ARGUMENT"
        elif status == FXNStatus.ERROR_INVALID_OPERATION:
            return "FXN_ERROR_INVALID_OPERATION"
        elif status == FXNStatus.ERROR_NOT_IMPLEMENTED:
            return "FXN_ERROR_NOT_IMPLEMENTED"
        return ""

PREDICTION_FIELDS = f"""
id
tag
type
configuration
resources {{
    type
    url
    name
}}
latency
error
logs
created
"""

_FXN_TO_NP_DTYPE = {
    FXNDtype.FLOAT16:   dtype("float16"),
    FXNDtype.FLOAT32:   dtype("float32"),
    FXNDtype.FLOAT64:   dtype("float64"),
    FXNDtype.INT8:      dtype("int8"),
    FXNDtype.INT16:     dtype("int16"),
    FXNDtype.INT32:     dtype("int32"),
    FXNDtype.INT64:     dtype("int64"),
    FXNDtype.UINT8:     dtype("uint8"),
    FXNDtype.UINT16:    dtype("uint16"),
    FXNDtype.UINT32:    dtype("uint32"),
    FXNDtype.UINT64:    dtype("uint64"),
    FXNDtype.BOOL:      dtype("bool"),
}

_NP_TO_FXN_DTYPE = { value: key for key, value in _FXN_TO_NP_DTYPE.items() }